
import 'package:clean_application/src/data/datasource.dart';
import 'package:clean_application/src/domain/repository.dart';
import 'package:clean_application/src/domain/usecase.dart';
import 'package:clean_application/src/presentation/configuration.dart';
import 'package:clean_application/src/presentation/controller.dart';
import 'package:clean_application/src/util/instance_provider.dart';
import 'package:flutter/material.dart';


/// This class is the root of the application
/// It register all the datasources, repositories, use cases and controllers.
/// The above order is the order in which the registration is done because of the
/// Clean architecture call flow.
/// The public static methods is used to access the registered instance on each
/// layer.
///
/// TODO: Add custom map for user custom data?
///
class CleanApplication extends InheritedModel {
  static Map<String, DataSource> _datasources = Map();
  static Map<String, Repository> _repositories = Map();
  static Map<String, UseCase> _usecases = Map();
  static Map<String, Controller> _controllers = Map();

  final String environment;
  final Configuration configuration;

  CleanApplication({
    this.environment = Environments.Development,
    @required this.configuration,
    @required Widget child,
  }) : assert(environment != null), assert(configuration != null), super(child: child) {

    /// This call order is important because of the layers dependency (call flow)
    _initializeInstanceProviderList(configuration.dataSourcesProvider(environment), _datasources);
    _initializeInstanceProviderList(configuration.repositoriesProvider(environment), _repositories);
    _initializeList(configuration.useCasesProvider(), _usecases);
    _initializeList(configuration.controllersProvider(), _controllers);
  }

  void _initializeInstanceProviderList(List<InstanceProvider> list, Map map){
    if(list == null){
      return;
    }

    for(var provider in list){
      _injectInMap(map, provider.instanciate(), provider.name());
    }
  }

  void _initializeList(List list, Map map) {
    if (list != null) {
      for (var val in list) {
        _injectInMap(map, val, '${val.runtimeType}');
      }
    }
  }

  static void _injectInMap<T>(Map map, T instance, [String name]) {
    if (name != null && name.isNotEmpty) {
      map[name] = instance;
    } else {
      map['$T'] = instance;
    }
  }

  static _getFromMap<T>(Map map, [String name]) {
    T value;
    if (name != null && name.isNotEmpty) {
      value = map[name] as T;
    } else {
      value = map['$T'] as T;
    }

    return value;
  }

  static datasource<T>([String name]) {
    return _getFromMap<T>(_datasources, name);
  }

  static repository<T>([String name]) {
    return _getFromMap<T>(_repositories, name);
  }

  static usecase<T>([String name]) {
    return _getFromMap<T>(_usecases, name);
  }

  static controller<T>([String name]) {
    return _getFromMap<T>(_controllers, name);
  }

  //static CleanApplication of1(BuildContext context) => context.dependOnInheritedWidgetOfExactType();

  static CleanApplication of(BuildContext context, {dynamic aspect}) {
    return InheritedModel.inheritFrom<CleanApplication>(context,
        aspect: aspect);
  }

  @override
  bool updateShouldNotify(InheritedWidget oldWidget) {
    return true;
  }

  @override
  bool updateShouldNotifyDependent(InheritedModel oldWidget, Set dependencies) {
    return true;
  }
}
